import 'dart:io';
import 'package:dio/dio.dart';
import 'package:flutter/foundation.dart';
import 'package:flutter/material.dart';

///## **BagelDB** is a content management system with rich API features. For more info, visit here bageldb.com

/// *BagelDB* class is the base class for any bagelDB request, it must be given the api [token] created at app.bageldb.com
class BagelDB {
  String token;
  BagelDB(this.token);

  /// return a *BagelDBRequest* object for a specific [collection]
  BagelDBRequest collection(String collection) {
    return BagelDBRequest(bagelDB: this, collection: collection);
  }
}

/// The base bagelDB response class, the response from bagelDB is inserted into the class
class BagelResponse {
  dynamic data;
  int itemCount;
  int statusCode;
  BagelResponse(
      {@required this.data, @required this.statusCode, this.itemCount});

  @override
  String toString() {
    String str = 'status code: ${this.statusCode.toString()}'
        'ֿ\nitem count: ${this.itemCount.toString()}'
        '\ndata: ${this.data.toString()}';
    return str;
  }
}

/// The builder for creating a bagelDB request, follow the builder pattern with the verbs get, put, post and delete building and executing the request. There is also the ability to use the uploadImage builder, for uploading images to an item
class BagelDBRequest {
  BagelDB bagelDB;
  String collection, itemID, sort, order;
  int pageNum, resPerPage;
  List<String> queryLst = [];
  bool callEverything;
  final String baseEndpoint = 'https://api.bagelstudio.co/api/public';
  Dio dio = Dio();

  BagelDBRequest(
      {@required this.bagelDB,
      @required this.collection,
      this.itemID,
      this.sort,
      this.order,
      this.resPerPage = 20,
      this.callEverything = false});

  /// Build and execute a get request to bagelDB, for both full collection and item requests
  Future<BagelResponse> get() async {
    var options = Options();
    options.headers['authorization'] = 'Bearer ${this.bagelDB.token}';
    options.headers["Accept-Version"] = "v1";

    Map<String, dynamic> params = Map<String, dynamic>();
    if (this.sort != null) params['sort'] = this.sort;
    if (this.order != null) params['order'] = this.order;
    if (this.callEverything)
      params['everything'] = this.callEverything.toString();
    if (this.pageNum != null) params['pageNumber'] = this.pageNum;
    if (this.resPerPage != null) params['perPage'] = this.resPerPage;
    String item = this.itemID != null ? '/' + this.itemID : '';
    String url = '$baseEndpoint/collection/${this.collection}/items$item';
    if (this.queryLst.isNotEmpty)
      url = url + "?query=" + this.queryLst.join("%2B");
    Response response =
        await dio.get(url, queryParameters: params, options: options);
    BagelResponse res = BagelResponse(
        data: response.data,
        statusCode: response.statusCode,
        itemCount: this.itemID != null
            ? 1
            : int.tryParse(response.headers['item-count'][0]));
    return res;
  }

  /// Build and execute a post request to bagelDB. [item] should follow the collection schema as defined at app.bageldb.com
  Future<BagelResponse> post({@required Map<String, dynamic> item}) async {
    var options = Options();
    options.headers['authorization'] = 'Bearer ${this.bagelDB.token}';
    options.headers["Accept-Version"] = "v1";

    String url = '$baseEndpoint/collection/${this.collection}/items';
    return dio.post(url, data: item, options: options).then((res) async {
      return BagelResponse(data: res.data, statusCode: res.statusCode);
    });
  }

  /// Build and execute a image upload to bagelDB, for a specific item in a collection. [image] should be a valid file object and [slug] is the field to which the image will be updated. The item() methods must have been used in order to execute uploadImage
  Future<BagelResponse> uploadImage(String slug, File image) async {
    var options = Options();
    options.headers['authorization'] = 'Bearer ${this.bagelDB.token}';
    options.headers["Accept-Version"] = "v1";

    String fileName = image.path.split('/').last;
    FormData formData = FormData.fromMap({
      "imageFile": await MultipartFile.fromFile(
        image.path,
        filename: fileName,
      ),
    });
    String url =
        '$baseEndpoint/collection/${this.collection}/items/${this.itemID}/image?imageSlug=$slug';
    Response res = await dio.put(url, data: formData, options: options);
    return BagelResponse(data: res.data, statusCode: res.statusCode);
  }

  /// Build and execute a put request to bagelDB. [item] should follow the collection schema as defined at app.bageldb.com
  Future<BagelResponse> put({@required Map<String, dynamic> item}) async {
    var options = Options();
    options.headers['authorization'] = 'Bearer ${this.bagelDB.token}';
    options.headers["Accept-Version"] = "v1";

    String url =
        '$baseEndpoint/collection/${this.collection}/items/${this.itemID}';
    Response res = await dio.put(url, data: item, options: options);
    return BagelResponse(data: res.data, statusCode: res.statusCode);
  }

  /// Build and execute a delete request to bagelDB.
  Future<BagelResponse> delete() async {
    var options = Options();
    options.headers['authorization'] = 'Bearer ${this.bagelDB.token}';
    options.headers["Accept-Version"] = "v1";

    String url =
        '$baseEndpoint/collection/${this.collection}/items/${this.itemID}';
    Response res = await dio.delete(url, options: options);
    return BagelResponse(data: res.data, statusCode: res.statusCode);
  }

  /// quering the collection with a [key] and [value] parameters, by the [queryOperator]. The following are valid queryOperators '=', '!=', 'regex', '>', '<'. The value can also be a comma seperated list of values, allowing for an 'in' query.
  /// For example: ```query('age', '=', '32')``` or for an itemRef query ```query('position.itemRefID', '=', '[432EWDWE]')```
  query(String key, String queryOperator, String value) {
    if (key == null) return this;
    String query = key + ":" + queryOperator + ":" + value;
    this.queryLst.add(Uri.encodeQueryComponent(query));
    return this;
  }

  /// sort data by [sortTerm] - a specific field, for example ```age```,
  /// and by [orderTerm] - ```desc``` for descending order, ```asc``` for ascending order.
  sortBy(String sortTerm, {String orderTerm}) {
    this.sort = sortTerm;
    if (orderTerm != null) this.order = orderTerm;
    return this;
  }

  /// define the [resultsPerPage] - how many results will be in each page
  perPage(int resultsPerPage) {
    this.resPerPage = resultsPerPage;
    return this;
  }

  /// define the [pageNumber] - the index of the page
  pageNumber(int pageNumber) {
    this.pageNum = pageNumber;
    return this;
  }

  /// defines the request for a specific item by the item's [_id]
  item(String _id) {
    this.itemID = _id;
    return this;
  }

  /// specify whether the request should include all the details for item references
  everything() {
    this.callEverything = true;
    return this;
  }
}
